// hub_mem

/*
-------------------------------------------------------------------------------
Copyright 2014 Parallax Inc.

This file is part of the hardware description for the Propeller 1 Design.

The Propeller 1 Design is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

The Propeller 1 Design is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
the Propeller 1 Design.  If not, see <http://www.gnu.org/licenses/>.
-------------------------------------------------------------------------------
*/
// RR20150509  optional code for DISABLE_FONT_ROM, CLUSO_ROMHI, NOT_SCRAMBLED 
// RR20150509  try $readmemh for rom(doesn't like intel hex)
// RR20150510  rom high as ram and byte format with ROM_HIGH_WRITABLE option
// RR20150511  rom low  as ram and byte format with ROM_FONT_WRITABLE option


module              hub_mem
(
input               clk_cog,
input               ena_bus,

input               w,
input        [3:0]  wb,
input       [13:0]  a,
input       [31:0]  d,

output      [31:0]  q
);


// 8192 x 32 ram with byte-write enables ($0000..$7FFF)

reg [7:0] ram3 [8191:0];
reg [7:0] ram2 [8191:0];
reg [7:0] ram1 [8191:0];
reg [7:0] ram0 [8191:0];

reg [7:0] ram_q3;
reg [7:0] ram_q2;
reg [7:0] ram_q1;
reg [7:0] ram_q0;

always @(posedge clk_cog)
begin
    if (ena_bus && !a[13] && w && wb[3])
        ram3[a[12:0]] <= d[31:24];
    if (ena_bus && !a[13])
        ram_q3 <= ram3[a[12:0]];
end

always @(posedge clk_cog)
begin
    if (ena_bus && !a[13] && w && wb[2])
        ram2[a[12:0]] <= d[23:16];
    if (ena_bus && !a[13])
        ram_q2 <= ram2[a[12:0]];
end

always @(posedge clk_cog)
begin
    if (ena_bus && !a[13] && w && wb[1])
        ram1[a[12:0]] <= d[15:8];
    if (ena_bus && !a[13])
        ram_q1 <= ram1[a[12:0]];
end

always @(posedge clk_cog)
begin
    if (ena_bus && !a[13] && w && wb[0])
        ram0[a[12:0]] <= d[7:0];
    if (ena_bus && !a[13])
        ram_q0 <= ram0[a[12:0]];
end


// 4096 x 32 rom containing character definitions ($8000..$BFFF)

`ifndef DISABLE_FONT_ROM
(* ram_init_file = "hub_rom_low_b3.hex" *)     reg [7:0] rom_low3 [4095:0];     // font
(* ram_init_file = "hub_rom_low_b2.hex" *)     reg [7:0] rom_low2 [4095:0];     // font
(* ram_init_file = "hub_rom_low_b1.hex" *)     reg [7:0] rom_low1 [4095:0];     // font
(* ram_init_file = "hub_rom_low_b0.hex" *)     reg [7:0] rom_low0 [4095:0];     // font

reg [7:0] rom_low_q3;
reg [7:0] rom_low_q2;
reg [7:0] rom_low_q1;
reg [7:0] rom_low_q0;

always @(posedge clk_cog)
begin
    `ifdef ROM_FONT_WRITABLE
    if (ena_bus && (a[13:12] == 2'b10) && w && wb[3])
        rom_low3[a[11:0]] <= d[31:24];
    `endif
    if (ena_bus && (a[13:12] == 2'b10))
        rom_low_q3 <= rom_low3[a[11:0]];
end

always @(posedge clk_cog)
begin
    `ifdef ROM_FONT_WRITABLE
    if (ena_bus && (a[13:12] == 2'b10) && w && wb[2])
        rom_low2[a[11:0]] <= d[23:16];
    `endif
    if (ena_bus && (a[13:12] == 2'b10))
        rom_low_q2 <= rom_low2[a[11:0]];
end

always @(posedge clk_cog)
begin
    `ifdef ROM_FONT_WRITABLE
    if (ena_bus && (a[13:12] == 2'b10) && w && wb[1])
        rom_low1[a[11:0]] <= d[15:8];
    `endif
    if (ena_bus && (a[13:12] == 2'b10))
        rom_low_q1 <= rom_low1[a[11:0]];
end

always @(posedge clk_cog)
begin
    `ifdef ROM_FONT_WRITABLE
    if (ena_bus && (a[13:12] == 2'b10) && w && wb[0])
        rom_low0[a[11:0]] <= d[7:0];
    `endif
    if (ena_bus && (a[13:12] == 2'b10))
        rom_low_q0 <= rom_low0[a[11:0]];
end
`endif

/*
`ifndef DISABLE_FONT_ROM
(* ram_init_file = "hub_rom_low.hex" *)     reg [7:0] rom_low3 [4095:0];

reg [31:0] rom_low_q;

always @(posedge clk_cog)
if (ena_bus && a[13:12] == 2'b10)
    rom_low_q <= rom_low[a[11:0]];
`endif
*/

// 4096 x 32 rom containing sin table, log table, booter, and interpreter ($C000..$FFFF)
//   .. with byte write enables (ie also works as ram)

`ifdef CLUSO_ROMHI
(* ram_init_file = "hub_rom_high_cluso_romhi_031_b3.hex" *)    reg [7:0] rom_high3 [4095:0];  // cluso interpreter unscrambled
(* ram_init_file = "hub_rom_high_cluso_romhi_031_b2.hex" *)    reg [7:0] rom_high2 [4095:0];  // cluso interpreter unscrambled
(* ram_init_file = "hub_rom_high_cluso_romhi_031_b1.hex" *)    reg [7:0] rom_high1 [4095:0];  // cluso interpreter unscrambled
(* ram_init_file = "hub_rom_high_cluso_romhi_031_b0.hex" *)    reg [7:0] rom_high0 [4095:0];  // cluso interpreter unscrambled
`elsif NOT_SCRAMBLED
(* ram_init_file = "hub_rom_high_unscrambled_b3.hex" *)        reg [7:0] rom_high3 [4095:0];  // unscrambled
(* ram_init_file = "hub_rom_high_unscrambled_b2.hex" *)        reg [7:0] rom_high2 [4095:0];  // unscrambled
(* ram_init_file = "hub_rom_high_unscrambled_b1.hex" *)        reg [7:0] rom_high1 [4095:0];  // unscrambled
(* ram_init_file = "hub_rom_high_unscrambled_b0.hex" *)        reg [7:0] rom_high0 [4095:0];  // unscrambled
`else
(* ram_init_file = "hub_rom_high_scrambled_b3.hex" *)          reg [7:0] rom_high3 [4095:0];  // scrambled
(* ram_init_file = "hub_rom_high_scrambled_b2.hex" *)          reg [7:0] rom_high2 [4095:0];  // scrambled
(* ram_init_file = "hub_rom_high_scrambled_b1.hex" *)          reg [7:0] rom_high1 [4095:0];  // scrambled
(* ram_init_file = "hub_rom_high_scrambled_b0.hex" *)          reg [7:0] rom_high0 [4095:0];  // scrambled
`endif

reg [7:0] rom_high_q3;
reg [7:0] rom_high_q2;
reg [7:0] rom_high_q1;
reg [7:0] rom_high_q0;

always @(posedge clk_cog)
begin
    `ifdef ROM_HIGH_WRITABLE
    if (ena_bus && (a[13:12] == 2'b11) && w && wb[3])
        rom_high3[a[11:0]] <= d[31:24];
    `endif
    if (ena_bus && (a[13:12] == 2'b11))
        rom_high_q3 <= rom_high3[a[11:0]];
end

always @(posedge clk_cog)
begin
    `ifdef ROM_HIGH_WRITABLE
    if (ena_bus && (a[13:12] == 2'b11) && w && wb[2])
        rom_high2[a[11:0]] <= d[23:16];
    `endif
    if (ena_bus && (a[13:12] == 2'b11))
        rom_high_q2 <= rom_high2[a[11:0]];
end

always @(posedge clk_cog)
begin
    `ifdef ROM_HIGH_WRITABLE
    if (ena_bus && (a[13:12] == 2'b11) && w && wb[1])
        rom_high1[a[11:0]] <= d[15:8];
    `endif
    if (ena_bus && (a[13:12] == 2'b11))
        rom_high_q1 <= rom_high1[a[11:0]];
end

always @(posedge clk_cog)
begin
    `ifdef ROM_HIGH_WRITABLE
    if (ena_bus && (a[13:12] == 2'b11) && w && wb[0])
        rom_high0[a[11:0]] <= d[7:0];
    `endif
    if (ena_bus && (a[13:12] == 2'b11))
        rom_high_q0 <= rom_high0[a[11:0]];
end



// 4096 x 32 rom containing sin table, log table, booter, and interpreter ($C000..$FFFF)

/*
`ifdef CLUSO_ROMHI
(* ram_init_file = "hub_rom_high_cluso_romhi_031.hex" *)    reg [31:0] rom_high [4095:0];  // cluso interpreter unscrambled
`elsif NOT_SCRAMBLED
(* ram_init_file = "hub_rom_high_unscrambled.hex" *)        reg [31:0] rom_high [4095:0];  // unscrambled
`else
(* ram_init_file = "hub_rom_high_scrambled.hex" *)          reg [31:0] rom_high [4095:0];  // scrambled
`endif
*/

/*
doesn't work! intel format!
reg [31:0] rom_high [4095:0];

initial
begin
`ifdef CLUSO_ROMHI
    $readmemh ("hub_rom_high_cluso_romhi_031.hex", rom_high);         // cluso interpreter unscrambled	
`elsif NOT_SCRAMBLED
    $readmemh ("hub_rom_high_unscrambled.hex", rom_high);             // unscrambled
`else
    $readmemh ("hub_rom_high_scrambled.hex", rom_high);               // scrambled
`endif
end
*/

/*
reg [31:0] rom_high_q;

always @(posedge clk_cog)
if (ena_bus && a[13:12] == 2'b11)
    rom_high_q <= rom_high[a[11:0]];
*/

// memory output mux

reg [1:0] mem;

always @(posedge clk_cog)
if (ena_bus)
    mem <= a[13:12];

assign q            = !mem[1]   ? {ram_q3, ram_q2, ram_q1, ram_q0}
`ifndef DISABLE_FONT_ROM
                    : !mem[0]   ? {rom_low_q3, rom_low_q2, rom_low_q1, rom_low_q0}     // comment out this line for DE0-Nano (sacrifices character rom to fit device)
`endif
                                : {rom_high_q3, rom_high_q2, rom_high_q1, rom_high_q0};

endmodule
